#include <math.h>
#include <stdlib.h>
#include "generate.h"
#include "world.h"
#include "voxels.h"
#include "heightmap.h"
#include "generate_utils.h"

void generate_flat(const unsigned int seed) {
	GU_seed(seed);

	skyHorizon = 0x9BCCE9;
	skyAtmosphere = 0x6AA8D3;
	skyZenith = 0x366AA3;
	sunColour = 0xFFFFFF;
	seaType = VOXEL_WATER;

	int x, y, z;
	for (x = 0; x < WORLDSIZE; x ++) {
		for (z = 0; z < WORLDSIZE; z ++) {
			for (y = 0; y < WORLDSIZE; y ++) {
				if (y < 1) {
					world_set(x, y, z, VOXEL_DIRT);
				} else if (y == 1) {
					world_set(x, y, z, VOXEL_GRASS);
				} else {
					world_set(x, y, z, 0);
				}
			}
		}
	}

	GU_undergrowth(1);
}

void generate_island(const unsigned int seed) {
	GU_seed(seed);

	skyHorizon = 0x9BCCE9;
	skyAtmosphere = 0x6AA8D3;
	skyZenith = 0x366AA3;
	sunColour = 0xFFFFFF;
	seaType = VOXEL_WATER;

	// Calculate heightmap, big blobby shape
	Heightmap base;
	size_t index;
	heightmap_clear(&base);
	for (index = 0; index < WORLDSIZE*WORLDSIZE/30; index ++) {
		heightmap_hump(&base, rand()%WORLDSIZE, rand()%WORLDSIZE, fabs(GU_noise())*8 + 2, fabs(GU_noise())*8 + 2);
	}
	heightmap_bias(&base);

	// Create stone with a layer of dirt that thins with altitude
	heightmap_blocks(&base, -8, 1, VOXEL_STONE);
	heightmap_blocks(&base, -3, 0.8, VOXEL_DIRT);

	// Make structure more interesting
	// NOTE: ore() can add empty places, so is called before calling GU_collapse()
	GU_ore(5);
	GU_collapse(VOXEL_DIRT);

	// Add life
	GU_undergrowth(1);
	const unsigned int foliage = (fabs(GU_noise()*2 + 0.2))*WORLDSIZE*WORLDSIZE/30;
	for (index = 0; index < foliage; index ++) {
		GU_tree(rand()%WORLDSIZE, rand()%WORLDSIZE);
	}

	// Add a layer of clouds
	GU_cloud(&base, rand()%(WORLDSIZE/2) + WORLDSIZE/2);
}

void generate_desert(const unsigned int seed) {
	GU_seed(seed);

	skyHorizon = 0xDCD7A7;
	skyAtmosphere = 0x9CD0DA;
	skyZenith = 0xF6F4E6;
	sunColour = 0xFFFFFF;
	seaType = VOXEL_SAND;

	// Calculate heightmap, lumpy blob
	Heightmap base;
	size_t index;
	heightmap_clear(&base);
	for (index = 0; index < WORLDSIZE*WORLDSIZE/5; index ++) {
		heightmap_hump(&base, rand()%WORLDSIZE, rand()%WORLDSIZE, fabs(GU_noise())*5 + 1, fabs(GU_noise())*2 + 1);
	}
	heightmap_bias(&base);

	// Create stone with a fairly flat layer of sand
	heightmap_blocks(&base, -8, 1, VOXEL_SANDSTONE);
	heightmap_blocks(&base, -3, 0.5, VOXEL_SAND);
	GU_collapse(VOXEL_SAND);

	// Add life
	const unsigned int foliage = fabs(GU_noise())*8 + 8;
	for (index = 0; index < foliage; index ++) {
		GU_cactus(rand()%WORLDSIZE, rand()%WORLDSIZE);
	}
	GU_undergrowth(1);
}

void generate_flight(const unsigned int seed) {
	GU_seed(seed);

	skyHorizon = 0x9BCCE9;
	skyAtmosphere = 0x6AA8D3;
	skyZenith = 0x366AA3;
	sunColour = 0xFFFFFF;
	seaType = VOXEL_CUMULUS;

	Densitymap base;
	size_t index;
	densitymap_clear(&base);
	for (index = 0; index < WORLDSIZE*WORLDSIZE*WORLDSIZE/20; index ++) {
		densitymap_blob(
			&base,
			rand()%WORLDSIZE,
			rand()%(WORLDSIZE*3/4),
			rand()%WORLDSIZE,
			fabs(GU_noise())*2 + 1,
			GU_noise() + 1);
	}
	densitymap_bias(&base);

	// Create stone with thin layer of grass
	densitymap_blocks(&base, 0, 1, 0, VOXEL_STONE);
	densitymap_blocks(&base, 0, 2, 2, VOXEL_DIRT);
	GU_stalactites(5, VOXEL_STONE);
	GU_ore(5);

	// Add life
	GU_undergrowth(1);
	const unsigned int foliage = fabs(GU_noise())*8 + 8;
	for (index = 0; index < foliage; index ++) {
		GU_tree(rand()%WORLDSIZE, rand()%WORLDSIZE);
	}
}

void generate_cave(const unsigned int seed) {
	GU_seed(seed);

	skyHorizon = 0x565A57;
	skyAtmosphere = 0x434443;
	skyZenith = 0x161616;
	sunColour = 0x808080;
	seaType = VOXEL_STONE;

	// Calculate heightmap, big blobby shape
	Heightmap base;
	size_t index;
	heightmap_clear(&base);
	for (index = 0; index < WORLDSIZE*WORLDSIZE/30; index ++) {
		heightmap_hump(&base, rand()%WORLDSIZE, rand()%WORLDSIZE, fabs(GU_noise())*5 + 1, fabs(GU_noise())*10 + 1);
	}
	heightmap_bias(&base);

	// Create stone with a sparse layer of sand that thins with altitude
	heightmap_blocks(&base, -1, 1, VOXEL_STONE);
	heightmap_blocks(&base, 0, 0.4, VOXEL_SAND);

	GU_ore(5);
	GU_stalactites(8, VOXEL_STONE);
	GU_undergrowth(0);

	// Mushrooms
	const unsigned int foliage = fabs(GU_noise())*8 + 8;
	for (index = 0; index < foliage; index ++) {
		GU_mushroom(rand()%WORLDSIZE, rand()%WORLDSIZE);
	}

	// Glowing slime
	int vx, vz;
	for (vx = 0; vx < WORLDSIZE; vx ++) {
		for (vz = 0; vz < WORLDSIZE; vz ++) {
			if (world_get(vx, 0, vz) == VOXEL_STONE && rand()%10 == 0) {
				world_set(vx, 0, vz, VOXEL_LAMP2);
			}
		}
	}
}

void generate_aether(const unsigned int seed) {
	GU_seed(seed);

	skyHorizon = 0x9BCCE9;
	skyAtmosphere = 0x6AA8D3;
	skyZenith = 0x366AA3;
	sunColour = 0xFFFFFF;
	seaType = VOXEL_CUMULUS;

	// Setup first heightmap
	Heightmap base;
	size_t index;
	heightmap_clear(&base);
	for (index = 0; index < WORLDSIZE*WORLDSIZE/40; index ++) {
		heightmap_hump(&base, rand()%WORLDSIZE, rand()%WORLDSIZE, fabs(GU_noise())*5 + 2, fabs(GU_noise())*5 + 2);
	}
	heightmap_bias(&base);

	// Add stratified cloud
	for (index = 0; index < WORLDSIZE; index += 5) {
		// Mutate heightmap
		heightmap_hump(&base, rand()%WORLDSIZE, rand()%WORLDSIZE, fabs(GU_noise())*5 + 2, fabs(GU_noise())*5 + 2);
		heightmap_hump(&base, rand()%WORLDSIZE, rand()%WORLDSIZE, fabs(GU_noise())*5 + 2, -fabs(GU_noise())*5 - 2);

		// Add clouds
		GU_cloud(&base, index + rand()%5);
	}

	// Dignity++
	int x, y, z;
	const unsigned int count = rand() % (WORLDSIZE*WORLDSIZE/20) + 1;
	unsigned short type;
	for (index = 0; index < count; index ++) {
		x = rand()%WORLDSIZE;
		z = rand()%WORLDSIZE;
		if (rand()%4 == 0) {
			type = VOXEL_WATER;
		} else {
			type = VOXEL_RAINBOW;
		}
		for (y = WORLDSIZE-1; y > 0; y --) {
			if (world_get(x, y, z) != 0 && world_get(x, (y-1), z) == 0) {
				for (; y > 0; y --) {
					world_set(x, y-1, z, type);
				}
				break;
			}
		}
	}
}

