#include <stdlib.h>
#include <math.h>
#include "generate_utils.h"
#include "world.h"
#include "voxels.h"

// State for GU_noise()
static float spare;
static char spareReady = 0;

/* Seed gaussian noise generator. */
void GU_seed(const unsigned int seed) {
	srand(seed);
	spareReady = 0;
}

float GU_noise() {
	if (spareReady) {
		spareReady = 0;
		return spare;
	}
	float s = 0;
	float u, v;
	do {
		u = (float)(rand()%1000)/500 * 2 - 1;
		v = (float)(rand()%1000)/500 - 1;
		s = u*u + v*v;
	} while (s >= 1 || s == 0);
	float mul = sqrt(-2*log(s)/s);
	spare = v*mul;
	spareReady = 1;
	return u*mul;
}

void GU_tree(const int vx, const int vz) {
	if (vx < 1 || vx >= WORLDSIZE-1 || vz < 1 || vz >= WORLDSIZE-1) {
		return;
	}

	const int trunk = fabs(GU_noise())*2 + 3;

	int ground;
	for (ground = WORLDSIZE-1; ground > 0; ground --) {
		if (solid[world_get(vx, ground, vz)]) {
			break;
		}
	}
	if (ground == 0 || ground >= WORLDSIZE-trunk*2 || world_get(vx, ground, vz) != VOXEL_GRASS) {
		return;
	}

	int x, y, z;
	for (y = ground; y < ground+trunk*2 && y < WORLDSIZE; y ++) {
		if (y == ground) {
			world_set(vx, y, vz, VOXEL_DIRT);
		} else if (y < ground+trunk) {
			world_set(vx, y, vz, VOXEL_WOOD);
		} else {
			for (x = vx-1; x <= vx+1; x ++) {
				for (z = vz-1; z <= vz+1; z ++) {
					world_set(x, y, z, VOXEL_LEAF);
				}
			}
		}
	}
}

void GU_mushroom(const int vx, const int vz) {
	const int trunk = fabs(GU_noise()) + 2;
	const int radius = fabs(GU_noise())*2 + 1;

	if (vx < radius || vx >= WORLDSIZE-radius || vz < radius || vz >= WORLDSIZE-radius) {
		return;
	}

	int ground;
	for (ground = 0; ground < WORLDSIZE-1; ground ++) {
		if (world_get(vx, ground+1, vz) == VOXEL_AIR) {
			break;
		}
	}
	if (ground == 0 || ground >= WORLDSIZE-trunk-2 || world_get(vx, ground, vz) != VOXEL_STONE) {
		return;
	}

	int x, y, z;
	for (y = ground; y <= ground+trunk+1; y ++) {
		if (y == ground) {
			world_set(vx, y, vz, VOXEL_STONE);
		} else if (y == ground+trunk) {
			for (x = -radius; x <= radius; x ++) {
				for (z = -radius; z <= radius; z ++) {
					if (x*x+z*z < radius*radius) {
						world_set(vx+x, y, vz+z, VOXEL_LARGEMUSHROOM);
					}
				}
			}
		} else if (y == ground+trunk+1) {
			for (x = -radius+1; x <= radius-1; x ++) {
				for (z = -radius+1; z <= radius-1; z ++) {
					if (x*x+z*z < (radius-1)*(radius-1)) {
						world_set(vx+x, y, z+vz, VOXEL_LARGEMUSHROOM);
					}
				}
			}
		} else {
			world_set(vx, y, vz, VOXEL_COLOUR1);
		}
	}
}

void GU_cactus(const int vx, const int vz) {
	const int size = fabs(GU_noise())*2 + 2;

	int ground;
	for (ground = WORLDSIZE-1; ground > 0; ground --) {
		if (solid[world_get(vx, ground, vz)]) {
			break;
		}
	}
	if (ground == 0 || ground >= WORLDSIZE-size-1 || world_get(vx, ground, vz) != VOXEL_SAND) {
		return;
	}

	int y;
	for (y = ground; y <= ground+size; y ++) {
		if (y == ground) {
			world_set(vx, y, vz, VOXEL_SAND);
		} else if (y == ground+size) {
			if (rand()%10 == 0) {
				world_set(vx, y, vz, VOXEL_ROSE);
			} else if (rand()%10 == 0) {
				world_set(vx, y, vz, VOXEL_DAISY);
			}
		} else {
			world_set(vx, y, vz, VOXEL_CACTUS);
		}
	}
}

void GU_stalactites(const int length, const unsigned short type) {
	int vx, vy, vz, end;
	for (vx = 0; vx < WORLDSIZE; vx ++) {
		for (vy = WORLDSIZE-2; vy >= 0; vy --) {
			for (vz = 0; vz < WORLDSIZE; vz ++) {
				if (world_get(vx, vy+1, vz) == type &&
						world_get(vx, vy, vz) == VOXEL_AIR) {
					end = vy - GU_noise()*length*0.5 + length;
					if (end < 0) {
						end = 0;
					}
					for (; vy >= end && world_get(vx, vy, vz) == VOXEL_AIR; vy --) {
						world_set(vx, vy, vz, type);
					}
				}
			}
		}
	}
}

void GU_undergrowth(const char sunEnabled) {
	int x, y, z;
	int stop;
	char sun;
	for (x = 0; x < WORLDSIZE; x ++) {
		for (z = 0; z < WORLDSIZE; z ++) {
			sun = sunEnabled;
			for (y = WORLDSIZE-1; y >= 0; y --) {
				if (sun) {
					if (world_get(x, y, z) == VOXEL_AIR) {
						if (y > 0 && world_get(x, y-1, z) == VOXEL_GRASS) {
							if (rand()%10 == 0) {
								world_set(x, y, z, VOXEL_LARGEGRASS + rand()%5);
							}
						} else if (y > 0 && world_get(x, y-1, z) == VOXEL_SAND) {
							if (rand()%20 == 0) {
								world_set(x, y, z, VOXEL_DEADBUSH);
							}
						}
					} else if (world_get(x, y, z) == VOXEL_DIRT) {
						world_set(x, y, z, VOXEL_GRASS);
						sun = 0;
					} else if (world_get(x, y, z) != VOXEL_AIR) {
						sun = 0;
					}
				} else {
					if (world_get(x, y, z) == VOXEL_AIR &&
							y > 0 && world_get(x, y-1, z) != VOXEL_AIR &&
							rand()%10 == 0) {
						world_set(x, y, z, VOXEL_MUSHROOM + rand()%2);
					} else if (world_get(x, y, z) == VOXEL_DIRT &&
							y > 0 && world_get(x, y-1, z) == VOXEL_AIR &&
							rand()%7 == 0) {
						stop = y - fabs(GU_noise()*3 + 5);
						if (stop < 0) {
							stop = 0;
						}
						for (y --; y >= stop && world_get(x, y, z) == VOXEL_AIR; y --) {
							world_set(x, y, z, VOXEL_STEM);
						}
					}
				}
			}
		}
	}
}

static void recurse_ore(const int x, const int y, const int z, const unsigned short type, const int length) {
	if (length <= 0 || world_get(x, y, z) != VOXEL_STONE) {
		return;
	}

	world_set(x, y, z, type);
	
	switch (rand()%6) {
		case 0:
			return recurse_ore(x-1, y, z, type, length-1);
		case 1:
			return recurse_ore(x+1, y, z, type, length-1);
		case 2:
			return recurse_ore(x, y-1, z, type, length-1);
		case 3:
			return recurse_ore(x, y+1, z, type, length-1);
		case 4:
			return recurse_ore(x, y, z-1, type, length-1);
		case 5:
			return recurse_ore(x, y, z+1, type, length-1);
	}
}
void GU_ore(const int seamLength) {
	int x, y, z;
	for (x = 0; x < WORLDSIZE; x ++) {
		for (y = 0; y < WORLDSIZE; y ++) {
			for (z = 0; z < WORLDSIZE; z ++) {
				if (world_get(x, y, z) == VOXEL_STONE) {
					if (rand()%32 == 0) {
						switch (rand()%8) {
							case 0:
								recurse_ore(x, y, z, VOXEL_HAEMATITE, rand()%seamLength + seamLength);
								break;
							case 1:
								recurse_ore(x, y, z, VOXEL_BORNITE, rand()%seamLength + seamLength);
								break;
							case 2:
								recurse_ore(x, y, z, VOXEL_GOLDNUGGET, rand()%seamLength + seamLength);
								break;
							case 3:
								recurse_ore(x, y, z, VOXEL_LAVA, rand()%seamLength + seamLength);
								break;
							default:
								recurse_ore(x, y, z, VOXEL_AIR, rand()%seamLength + seamLength);
								break;
						}
					}
				}
			}
		}
	}
}

// Set *low, *high to bounding indices for axis of sphere
// NOTE: Useful because world is a cube
static void spherebox(const float centre, const float radius, int *low, int *high) {
	*low = fmax(floor(centre - radius), 0);
	*high = fmin(ceil(centre + radius), WORLDSIZE - 1);
}

void GU_sphere(const float x, const float y, const float z, const float radius, const unsigned short type) {
	if (radius < 0.5) return;

	// Calculate bounding box
	int xl, xh, yl, yh, zl, zh;
	spherebox(x, radius, &xl, &xh);
	spherebox(y, radius, &yl, &yh);
	spherebox(z, radius, &zl, &zh);
	
	// Add sphere
	int u, v, w;
	for (u = xl; u <= xh; u ++) {
		for (v = yl; v <= yh; v ++) {
			for (w = zl; w <= zh; w ++) {
				if (world_get(u, v, w) == VOXEL_AIR && pow(u - x, 2) + pow(v - y, 2) + pow(w - z, 2) < radius*radius) {
					world_set(u, v, w, type);
				}
			}
		}
	}
}

void GU_cloud(const Heightmap *data, const int y) {
	int x, z;
	float radius;
	for (x = 0; x < WORLDSIZE; x ++) {
		for (z = 0; z < WORLDSIZE; z ++) {
			radius = fmin(GU_noise()*2*heightmap_get(data, x, z)/WORLDSIZE, 3);
			GU_sphere(x, y, z, radius, VOXEL_CUMULUS);
		}
	}
}

void GU_collapse(const unsigned short type) {
	int x, y, z;
	for (x = 0; x < WORLDSIZE; x ++) {
		for (z = 0; z < WORLDSIZE; z ++) {
			for (y = 1; y < WORLDSIZE; y ++) {
				if (world_get(x, y, z) == type && world_get(x, y-1, z) == VOXEL_AIR) {
					world_set(x, y-1, z, type);
					world_set(x, y, z, VOXEL_AIR);
				}
			}
		}
	}
}

