#include <stdlib.h>
#include <math.h>
#include <SDL2/SDL.h>
#include "heightmap.h"
#include "world.h"
#include "voxels.h"

// Return index from coordinates
#define HEIGHTMAP_INDEX(x, y) ((x)*WORLDSIZE + y)
#define DENSITYMAP_INDEX(x, y, z) ((x)*WORLDSIZE*WORLDSIZE + (y)*WORLDSIZE + z)

void heightmap_clear(Heightmap *data) {
	size_t index;
	for (index = 0; index < WORLDSIZE*WORLDSIZE; index ++) {
		data->data[index] = 0;
	}
}

void heightmap_hump(Heightmap *data, const float x, const float y, const float radius, const float height) {
	// Check centre is within data
	SDL_assert(x >= 0 && x < WORLDSIZE);
	SDL_assert(y >= 0 && y < WORLDSIZE);

	// Calculate bounding box
	const int xl = fmax(floor(x - radius), 0);
	const int xh = fmin(ceil(x + radius), WORLDSIZE-1);
	const int yl = fmax(floor(y - radius), 0);
	const int yh = fmin(ceil(y + radius), WORLDSIZE-1);

	// Add heights
	int u, v;
	float delta;
	for (u = xl; u <= xh; u ++) {
		for (v = yl; v <= yh; v ++) {
			delta = sqrt(pow(u - x, 2) + pow(v - y, 2));
			if (delta < radius) {
				data->data[HEIGHTMAP_INDEX(u, v)] += (cos(delta/radius*M_PI)*0.5 + 0.5)*height;
			}
		}
	}
}

void heightmap_bias(Heightmap *data) {
	const int cx = WORLDSIZE/2;
	const int cy = WORLDSIZE/2;

	int x, y;
	float delta;
	for (x = 0; x < WORLDSIZE; x ++) {
		for (y = 0; y < WORLDSIZE; y ++) {
			delta = sqrt(pow(x - cx, 2) + pow(y - cy, 2));
			if (delta < WORLDSIZE/2) {
				data->data[HEIGHTMAP_INDEX(x, y)] *= cos(delta*2/WORLDSIZE*M_PI)*0.5 + 0.5;
			} else {
				data->data[HEIGHTMAP_INDEX(x, y)] = 0;
			}
		}
	}
}

void heightmap_blocks(const Heightmap *data, const float add, const float multiply, const unsigned short type) {
	int x, y, z;
	for (x = 0; x < WORLDSIZE; x ++) {
		for (z = 0; z < WORLDSIZE; z ++) {
			for (y = 0; y < WORLDSIZE && y < data->data[HEIGHTMAP_INDEX(x, z)]*multiply + add; y ++) {
				if (world_get(x, y, z) == VOXEL_AIR) {
					world_set(x, y, z, type);
				}
			}
		}
	}
}

float heightmap_get(const Heightmap *data, const int x, const int y) {
	return data->data[HEIGHTMAP_INDEX(x, y)];
}

void densitymap_clear(Densitymap *data) {
	size_t index;
	for (index = 0; index < WORLDSIZE*WORLDSIZE*WORLDSIZE; index ++) {
		data->data[index] = 0;
	}
}

void densitymap_blob(Densitymap *data, const float x, const float y, const float z, const float radius, const float weight) {
	// Check centre is within data
	SDL_assert(x >= 0 && x < WORLDSIZE);
	SDL_assert(y >= 0 && y < WORLDSIZE);
	SDL_assert(z >= 0 && z < WORLDSIZE);

	// Calculate bounding box
	const int xl = fmax(floor(x - radius), 0);
	const int xh = fmin(ceil(x + radius), WORLDSIZE-1);
	const int yl = fmax(floor(y - radius), 0);
	const int yh = fmin(ceil(y + radius), WORLDSIZE-1);
	const int zl = fmax(floor(z - radius), 0);
	const int zh = fmin(ceil(z + radius), WORLDSIZE-1);

	// Add heights
	int u, v, w;
	float delta;
	for (u = xl; u <= xh; u ++) {
		for (v = yl; v <= yh; v ++) {
			for (w = zl; w <= zh; w ++) {
				delta = sqrt(pow(u - x, 2) + pow(v - y, 2) + pow(w - z, 2));
				if (delta < radius) {
					data->data[DENSITYMAP_INDEX(u, v, w)] += (cos(delta/radius*M_PI)*0.5 + 0.5)*weight;
				}
			}
		}
	}
}

void densitymap_bias(Densitymap *data) {
	const int cx = WORLDSIZE/2;
	const int cy = WORLDSIZE/2;
	const int cz = WORLDSIZE/2;

	int x, y, z;
	float delta;
	for (x = 0; x < WORLDSIZE; x ++) {
		for (y = 0; y < WORLDSIZE; y ++) {
			for (z = 0; z < WORLDSIZE; z ++) {
				delta = sqrt(pow(x - cx, 2) + pow(y - cy, 2) + pow(z - cz, 2));
				if (delta < WORLDSIZE/2) {
					data->data[DENSITYMAP_INDEX(x, y, z)] *= cos(delta*2/WORLDSIZE*M_PI)*0.5 + 0.5;
				} else {
					data->data[DENSITYMAP_INDEX(x, y, z)] = 0;
				}
			}
		}
	}
}

void densitymap_blocks(
		const Densitymap *data,
		const float add,
		const float multiply,
		const int offset,
		const unsigned short type) {

	int x, y, z;
	float value;
	for (x = 0; x < WORLDSIZE; x ++) {
		for (y = 0; y < WORLDSIZE; y ++) {
			for (z = 0; z < WORLDSIZE; z ++) {
				if (y - offset < 0 || y - offset >= WORLDSIZE) {
					value = add;
				} else {
					value = data->data[DENSITYMAP_INDEX(x, y-offset, z)]*multiply + add;
				}

				if (world_get(x, y, z) == VOXEL_AIR && value > 1) {
					world_set(x, y, z, type);
				}
			}
		}
	}
}

