#include "config.h"
#include "player.h"
#include "resource.h"
#include "render.h"
#include "world.h"
#include "voxels.h"
#include "physics.h"

GLfloat cameraX, cameraY, cameraZ;
GLfloat cameraP, cameraT;

float velocityX, velocityY, velocityZ;

Mix_Chunk *sfx[VOXELTYPES];

char selected;
int selectX, selectY, selectZ;
char faceX, faceY, faceZ;

unsigned short player_tool;

// True to switch to flying
char fly = 1;

/* Type of voxel under character. */
unsigned short ground = 0;

/* Footstep sound effect timer. */
int footstep = 0;
#ifdef SPP
#define FOOTSTEPDELAY 12
#define FOOTSTEPDELAY2 5
unsigned char footstate = 0;
#else
#define FOOTSTEPDELAY 15
#endif

void player_init() {
	/* Clear sfx. */
	int i;
	for (i = 0; i < VOXELTYPES; i ++) {
		sfx[i] = NULL;
	}
}

void player_clear() {
	cameraX = WORLDSIZE/2+0.5;
	cameraY = WORLDSIZE+PLAYERBOTTOM;
	cameraZ = WORLDSIZE/2+0.5;
	cameraP = 0;
	cameraT = 0;
	player_tool = 1;
}

void player_look(const float dx, const float dy) {
	cameraP += dx;
	cameraT += dy;
	if (cameraP < -180) {
		cameraP += 360;
	}
	if (cameraP > 180) {
		cameraP -= 360;
	}
	if (cameraT < -90) {
		cameraT = -90;
	}
	if (cameraT > 90) {
		cameraT = 90;
	}
}

/* True when player is intersecting a solid voxel.
Sets *type to the first type of solid voxel found unless type is NULL. */
static int player_check(unsigned short *type) {
	int left = floor(cameraX-PLAYERRADIUS);
	int right = floor(cameraX+PLAYERRADIUS);
	int top = floor(cameraY+PLAYERTOP);
	int bottom = floor(cameraY-PLAYERBOTTOM);
	int front = floor(cameraZ+PLAYERRADIUS);
	int back = floor(cameraZ-PLAYERRADIUS);

	int x, y, z;
	for (x = left; x <= right; x ++) {
		for (y = bottom; y <= top; y ++) {
			for (z = back; z <= front; z ++) {
				if (x < 0 || x >= WORLDSIZE || y < 0 || y >= WORLDSIZE || z < 0 || z >= WORLDSIZE) {
					continue;
				}
				if (solid[voxels[x*WORLDSIZE*WORLDSIZE+y*WORLDSIZE+z]]) {
					if (type != NULL) {
						*type = voxels[x*WORLDSIZE*WORLDSIZE+y*WORLDSIZE+z];
					}
					return 1;
				}
			}
		}
	}
	return 0;
}

void player_move(float dx, float dy, float dz) {
	if (fly) {
		dx *= 0.01;
		dy *= 0.01;
		dz *= 0.01;
		velocityX += cos(cameraP*M_PI/180)*dx + sin(cameraP*M_PI/180)*dz;
		velocityY += dy;
		velocityZ += sin(cameraP*M_PI/180)*dx - cos(cameraP*M_PI/180)*dz;
	} else {
		if (velocityY == 0) {
			// Standing on solid voxel
			dx *= 0.005;
			dz *= 0.005;

			// Player footstep
			if (footstep == 0) {
				if (dx != 0 || dz != 0) {
					#ifdef SPP
					footstate = !footstate;
					if (footstate) {
						footstep = FOOTSTEPDELAY;
					} else {
						footstep = FOOTSTEPDELAY2;
					}
					#else
					footstep = FOOTSTEPDELAY;
					#endif
					if (sfx[ground] != NULL) {
						Mix_PlayChannel(-1, sfx[ground], 0);
					}
				}
			} else {
				footstep --;
			}
		} else {
			// Falling
			dx *= 0.001;
			dz *= 0.001;
		}
		dy *= 0.02;
		velocityX += cos(cameraP*M_PI/180)*dx + sin(cameraP*M_PI/180)*dz;
		if (velocityY == 0) {
			velocityY = dy;
		}
		velocityZ += sin(cameraP*M_PI/180)*dx - cos(cameraP*M_PI/180)*dz;
	}
}

void player_destroy() {
	// Get type of voxel
	unsigned short type = VOXEL_AIR;
	if (selected && selectX >= 0 && selectX < WORLDSIZE && selectY >= 0 && selectY < WORLDSIZE && selectZ >= 0 && selectZ < WORLDSIZE) {
		type = voxels[selectX*WORLDSIZE*WORLDSIZE+selectY*WORLDSIZE+selectZ];
	}

	if (physics_tool_destroyfn[player_tool] != NULL) {
		physics_tool_destroyfn[player_tool](type, selectX, selectY, selectZ, faceX, faceY, faceZ);
	} else if (type != VOXEL_AIR) {
		if (physics_destroyfn[type] == NULL || physics_destroyfn[type](type, selectX, selectY, selectZ, faceX, faceY, faceZ)) {
			// Player block destroy sound
			if (sfx[type] != NULL) {
				Mix_PlayChannel(-1, sfx[type], 0);
			}

			// Update world
			voxels[selectX*WORLDSIZE*WORLDSIZE+selectY*WORLDSIZE+selectZ] = 0;
			world_dirty();
			render_update();

			// Trigger physics
			physics_changed(selectX, selectY, selectZ);
		}
	}
}

void player_select() {
	// Get type of voxel
	unsigned short type = VOXEL_AIR;
	if (selected && selectX >= 0 && selectX < WORLDSIZE && selectY >= 0 && selectY < WORLDSIZE && selectZ >= 0 && selectZ < WORLDSIZE) {
		type = voxels[selectX*WORLDSIZE*WORLDSIZE+selectY*WORLDSIZE+selectZ];
	}

	if (physics_tool_selectfn[player_tool] != NULL) {
		physics_tool_selectfn[player_tool](type, selectX, selectY, selectZ, faceX, faceY, faceZ);
	} else if (type != VOXEL_AIR) {
		player_tool = voxels[selectX*WORLDSIZE*WORLDSIZE+selectY*WORLDSIZE+selectZ];
	}
}

void player_create() {
	if (physics_tool_createfn[player_tool] != NULL) {
		// Get type of voxel
		unsigned short type = VOXEL_AIR;
		if (selected && selectX >= 0 && selectX < WORLDSIZE && selectY >= 0 && selectY < WORLDSIZE && selectZ >= 0 && selectZ < WORLDSIZE) {
			type = voxels[selectX*WORLDSIZE*WORLDSIZE+selectY*WORLDSIZE+selectZ];
		}

		// Use tool
		physics_tool_createfn[player_tool](type, selectX, selectY, selectZ, faceX, faceY, faceZ);
	} else {
		int x, y, z;
		x = selectX + faceX;
		y = selectY + faceY;
		z = selectZ + faceZ;
		if (selected && x >= 0 && x < WORLDSIZE && y >= 0 && y < WORLDSIZE && z >= 0 && z < WORLDSIZE) {
			// Check if block intersects with character
			if (x+1 > cameraX-PLAYERRADIUS && x < cameraX+PLAYERRADIUS &&
					y+1 > cameraY-PLAYERBOTTOM && y < cameraY+PLAYERTOP &&
					z+1 > cameraZ-PLAYERRADIUS && z < cameraZ+PLAYERRADIUS) {
				return;
			}

			// Player block place sound
			if (sfx[player_tool] != NULL) {
				Mix_PlayChannel(-1, sfx[player_tool], 0);
			}

			// Update world
			voxels[x*WORLDSIZE*WORLDSIZE + y*WORLDSIZE + z] = player_tool;
			world_dirty();
			render_update();

			// Trigger physics
			physics_changed(x, y, z);
		}
	}
}

void player_activate() {
	unsigned short type;
	if (selected && selectX >= 0 && selectX < WORLDSIZE && selectY >= 0 && selectY < WORLDSIZE && selectZ >= 0 && selectZ < WORLDSIZE) {
		type = voxels[selectX*WORLDSIZE*WORLDSIZE + selectY*WORLDSIZE + selectZ];
		if(physics_activatefn[type] != NULL) {
			physics_activatefn[type](type, selectX, selectY, selectZ, faceX, faceY, faceZ);
		}
	}
}

void player_switch(const int add) {
	if ((int)player_tool + add <= 0) {
		player_tool = 1;
	} else if ((int)player_tool + add >= LOADEDVOXELS) {
		player_tool = LOADEDVOXELS - 1;
	} else {
		player_tool += add;
	}
}

void player_fly() {
	fly = !fly;
}

void player_update() {
	int i;
	// Iterate to allow higher velocities
	for (i = 0; i < 20; i ++) {
		// Add gravity to velocity
		if (fly) {
			velocityX *= 0.99;
			velocityY *= 0.99;
			velocityZ *= 0.99;
		} else {
			if (velocityY == 0) {
				velocityX *= 0.96;
				velocityZ *= 0.96;
			} else {
				velocityX *= 0.99;
				velocityZ *= 0.99;
			}
			if (cameraY > 0.2) {
				velocityY -= GRAVITY;
			}
		}

		// Add buoyancy to velocity
		if (!solid[seaType]) {
			if (cameraY < 0.2) {
				velocityY *= 0.99;
				if (cameraY < 0.2-0.0001) {
					velocityY += 0.0001;
				}
			}
		}

		// Limit velocity
		if (velocityX < -VELOCITYMAX) {
			velocityX = -VELOCITYMAX;
		} else if (velocityX > VELOCITYMAX) {
			velocityX = VELOCITYMAX;
		}
		if (velocityY < -VELOCITYMAX) {
			velocityY = -VELOCITYMAX;
		} else if (velocityY > VELOCITYMAX) {
			velocityY = VELOCITYMAX;
		}
		if (velocityZ < -VELOCITYMAX) {
			velocityZ = -VELOCITYMAX;
		} else if (velocityZ > VELOCITYMAX) {
			velocityZ = VELOCITYMAX;
		}

		// Apply velocity
		cameraX += velocityX;
		if (player_check(NULL)) {
			cameraX -= velocityX;
			velocityX = 0;
		}
		cameraY += velocityY;
		if (player_check(&ground)) {
			if (fabs(velocityY) > 0.01) {
				footstep = FOOTSTEPDELAY;
				Mix_PlayChannel(-1, sfx[ground], 0);
			}
			cameraY -= velocityY;
			velocityY = 0;
		}
		cameraZ += velocityZ;
		if (player_check(NULL)) {
			cameraZ -= velocityZ;
			velocityZ = 0;
		}

		// Check for intersection with floor if sea is solid {
		if (solid[seaType]) {
			if (cameraY < PLAYERBOTTOM) {
				ground = seaType;
				if (fabs(velocityY) > 0.01) {
					footstep = FOOTSTEPDELAY;
					Mix_PlayChannel(-1, sfx[ground], 0);
				}
				cameraY = PLAYERBOTTOM;
				velocityY = 0;
			}
		}
	}

	// Get selected voxel
	float cose = cos(cameraT*M_PI/180);
	selected = world_ray(cameraX, cameraY, cameraZ,
		sin(cameraP*M_PI/180)*cose, -sin(cameraT*M_PI/180), -cos(cameraP*M_PI/180)*cose,
		&selectX, &selectY, &selectZ,
		&faceX, &faceY, &faceZ);
}
