/* Provides functions for managing the world. */

#ifndef __WORLD_H__
#define __WORLD_H__

#include "config.h"

/* Voxel types to allocate. The engine supports up to 2**sizeof(unsigned short) = 65536. */
#define VOXELTYPES 255
/* Extra shadow added for every lighting step through transparent voxels. */
#define AIRSHADOW 16

/* Voxels. */
extern unsigned short voxels[];
/* Light. */
#ifndef DISABLELIGHTING
extern unsigned char light[WORLDSIZE*WORLDSIZE*WORLDSIZE*3];
/* Properties of types of voxel. */
// Amount of light absorbed by the voxel as an RGB colour
extern unsigned long shadow[VOXELTYPES];
// Amount of light emitted by the voxel as an RGB colour
extern unsigned long glow[VOXELTYPES];
#endif
// True if voxel is opaque (rendering is optimised by not drawing voxels that are hidden by opaque voxels)
// The texture of the voxel should be completely opaque
extern char opaque[];
// True if the voxel is solid (used for physics)
extern char solid[];
// Colour used for suffocation
extern unsigned long insideColour[VOXELTYPES];

/* Colour info. */
// At horizon
extern unsigned long skyHorizon;
// Over horizon
extern unsigned long skyAtmosphere;
// Zenith
extern unsigned long skyZenith;
// Colour ambient light
extern unsigned long sunColour;

/* Type of voxel the sea is made of. */
extern unsigned short seaType;

/* Setup world. */
void world_init();
/* Reset world. */
void world_clear();
/* Load world, return true if successful. */
char world_load(const char *path);
/* Save world, return true if successful. */
char world_save(const char *path);
/* Calculate lighting. */
void world_relight();
/* Queue lighting recalculation. */
void world_dirty();
/* Find the first solid voxel intersected by a ray.
Returns true if a voxel was hit.
hitx, hity, hitz will be overwritten with the position of the voxel that was hit.
facex, facey, facez will be overwritten with the direction of the face that was hit relative to the voxel. */
char world_ray(
	const float startx, const float starty, const float startz,
	const float dx, const float dy, const float dz,
	int *hitx, int *hity, int *hitz,
	char *facex, char *facey, char *facez);

/* Macros to get voxel and light indicies. */
#define WORLD_INDEX(x, y, z) ((x)*WORLDSIZE*WORLDSIZE + (y)*WORLDSIZE + (z))
#define WORLD_LIGHT(x, y, z) (WORLD_INDEX(x, y, z) * 3)

/* Return voxel at x, y, z.
NOTE: Bounds not checked!
NOTE: Implemented as a macro for speed */
#define world_get(x, y, z) voxels[WORLD_INDEX(x, y, z)]

/* Set voxel at x, y, z.
NOTE: Bounds not checked! */
#define world_set(x, y, z, type) voxels[WORLD_INDEX(x, y, z)] = type


#endif
